@@ -14,11 +14,13 @@ from typing import (
     TypeVar,
     Union,
     overload,
+    Hashable,
 )
 
 SLICE_ALL = slice(None)
 
 T = TypeVar("T")
+S = TypeVar("S", bound="StableSet")
 
 # SetLike[T] is either a set of elements of type T, or a sequence, which
 # we will convert to a StableSet or to an OrderedSet by adding its elements in order.
@@ -211,7 +213,7 @@ class StableSet(MutableSet[T], Sequence[T]):
     # (left hand and right hand - as the operands order does matter)
     # based on the implementations of the super class (Set(Collection)),
     # see _collections_abc.py
-    def __sub__(self, other):
+    def __sub__(self: S, other: AbstractSet[T]) -> S:
         cls = type(
             self
             if isinstance(self, StableSet)
@@ -225,7 +227,7 @@ class StableSet(MutableSet[T], Sequence[T]):
             other = cls(other)
         return cls(value for value in self if value not in other)
 
-    def __rsub__(self, other):
+    def __rsub__(self: S, other: AbstractSet[T]) -> S:
         cls = type(
             self
             if isinstance(self, StableSet)
@@ -240,7 +242,7 @@ class StableSet(MutableSet[T], Sequence[T]):
         return cls(value for value in other if value not in self)
 
 
-    def __or__(self, other):
+    def __or__(self: S, other: AbstractSet[T]) -> S:
         cls = type(
             self
             if isinstance(self, StableSet)
@@ -253,7 +255,7 @@ class StableSet(MutableSet[T], Sequence[T]):
         chain = (e for s in (self, other) for e in s)
         return cls(chain)
 
-    def __ror__(self, other):
+    def __ror__(self: S, other: AbstractSet[T]) -> S:
         cls = type(
             self
             if isinstance(self, StableSet)
@@ -266,17 +268,17 @@ class StableSet(MutableSet[T], Sequence[T]):
         chain = (e for s in (other, self) for e in s)
         return cls(chain)
 
-    def __xor__(self, other):
+    def __xor__(self: S, other: AbstractSet[T]) -> S:
         if not isinstance(other, Iterable):
             return NotImplemented
         return (self - other) | (other - self)
 
-    def __rxor__(self, other):
+    def __rxor__(self: S, other: AbstractSet[T]) -> S:
         if not isinstance(other, Iterable):
             return NotImplemented
         return (other - self) | (self - other)
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
         if not isinstance(other, Iterable):
             return False
         if len(self._map) != len(other):
@@ -313,7 +315,7 @@ class StableSet(MutableSet[T], Sequence[T]):
     # Technically type-incompatible with MutableSet, because we return an
     # int instead of nothing. This is also one of the things that makes
     # StableSet convenient to use.
-    def add(self, key: T) -> int:
+    def add(self, key: T) -> int:  # pyright: ignore
         """
         Add `key` as an item to this StableSet, then return its index.
 
@@ -355,15 +357,15 @@ class StableSet(MutableSet[T], Sequence[T]):
         return len(self._map) - 1
 
     @overload
-    def index(self, key: Sequence[T]) -> List[int]:  # NOQA
+    def index(self, key: Sequence[T]) -> List[int]:
         ...
 
     @overload
-    def index(self, key: T) -> int:  # NOQA
+    def index(self, key: T) -> int:
         ...
 
     # concrete implementation
-    def index(self, key):  # NOQA
+    def index(self, key: Union[T, Sequence[T]]) -> Union[int, List[int]]:  # pyright: ignore
         """
         Get the index of a given entry, raising an IndexError if it's not present
 
@@ -443,7 +445,7 @@ class StableSet(MutableSet[T], Sequence[T]):
         self._map.pop(key)
         self._map[key] = None
 
-    def discard(self, key: T) -> None:
+    def discard(self, value: T) -> None:
         """
         Remove an element.  Do not raise an exception if absent.
 
@@ -462,7 +464,7 @@ class StableSet(MutableSet[T], Sequence[T]):
         if self._is_mutable is False:
             raise ValueError("This object is not mutable.")
 
-        self._map.pop(key, None)
+        self._map.pop(value, None)
 
     def union(self, *sets: SetLike[T]) -> "StableSet[T]":
         """
@@ -483,7 +485,7 @@ class StableSet(MutableSet[T], Sequence[T]):
         items = it.chain.from_iterable(containers)
         return cls(items)  # type: ignore
 
-    def intersection(self, *sets: SetLike[T]) -> "StableSet[T]":
+    def intersection(self: S, *sets: SetLike[T]) -> S:
         """
         Returns elements in common between all sets. Order is defined only
         by the first set.
@@ -504,7 +506,7 @@ class StableSet(MutableSet[T], Sequence[T]):
             items = (item for item in self if item in common)
         return cls(items)
 
-    def difference(self, *sets: SetLike[T]) -> "StableSet[T]":
+    def difference(self: S, *sets: SetLike[T]) -> S:
         """
         Returns all elements that are in this set but not the others.
 
@@ -525,7 +527,7 @@ class StableSet(MutableSet[T], Sequence[T]):
             items = (item for item in self if item not in other)
         return cls(items)
 
-    def symmetric_difference(self, other: SetLike[T]) -> "StableSet[T]":
+    def symmetric_difference(self: S, other: SetLike[T]) -> S:
         """
         Return the symmetric difference of two StableSets as a new set.
         That is, the new set will contain all elements that are in exactly
@@ -577,7 +579,7 @@ class StableSet(MutableSet[T], Sequence[T]):
             [item for item in self._map if item not in items_to_remove]
         )
 
-    def intersection_update(self, other: SetLike[T]) -> None:
+    def intersection_update(self, other: SetLike[T]) -> T:
         """
         Update this StableSet to keep only items in another set, preserving
         their order in this set.
@@ -586,6 +588,7 @@ class StableSet(MutableSet[T], Sequence[T]):
             >>> this = StableSet([1, 4, 3, 5, 7])
             >>> other = StableSet([9, 7, 1, 3, 2])
             >>> this.intersection_update(other)
+            StableSet([1, 3, 7])
             >>> print(this)
             StableSet([1, 3, 7])
         """
@@ -593,6 +596,9 @@ class StableSet(MutableSet[T], Sequence[T]):
             raise ValueError("This object is not mutable.")
         other = set(other)
         self._map = dict.fromkeys([item for item in self._map if item in other])
+        return self
+
+    __iand__ = intersection_update
 
     def symmetric_difference_update(self, other: SetLike[T]) -> None:
         """
@@ -606,6 +612,8 @@ class StableSet(MutableSet[T], Sequence[T]):
             >>> print(this)
             StableSet([4, 5, 9, 2])
         """
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
         items_to_add = [item for item in other if item not in self]
         items_to_remove = set(other)
         self._map = dict.fromkeys(
@@ -644,7 +652,7 @@ class StableSet(MutableSet[T], Sequence[T]):
             return False
         return all(item in self for item in other)
 
-    def isorderedsubset(self: SetLike, other: SetLike, non_consecutive: bool = False):
+    def isorderedsubset(self: SetLike, other: SetLike, non_consecutive: bool = False) -> bool:
         if len(self) > len(other):
             return False
         if non_consecutive:
@@ -662,13 +670,13 @@ class StableSet(MutableSet[T], Sequence[T]):
                     return False
             return True
 
-    def isorderedsuperset(self, other: SetLike, non_consecutive: bool = False):
+    def isorderedsuperset(self, other: SetLike, non_consecutive: bool = False) -> bool:
         return StableSet.isorderedsubset(other, self, non_consecutive)
 
-    def get(self):
+    def get(self) -> Hashable:
         return next(iter(self._map))
 
-    def freeze(self):
+    def freeze(self) -> None:
         """
         Once this function is run, the object becomes immutable
         """
@@ -839,13 +847,15 @@ class OrderedSet(StableSet[T]):
         OrderedSet([1, 2, 3])
     """
 
-    __slots__ = ("_items",)
+    __slots__ = ("_items", "_is_mutable")
 
     _items: List[T]
 
     def __init__(self, initial: Optional[SetInitializer[T]] = None):
         self._items = []
         self._map = {}
+        self._is_mutable = True
+
         if initial is not None:
             # In terms of duck-typing, the default __ior__ is compatible with
             # the types we use, but it doesn't expect all the types we
@@ -933,16 +943,25 @@ class OrderedSet(StableSet[T]):
         )
 
     def clear(self) -> None:
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
+
         del self._items[:]
         self._map.clear()
 
     def add(self, key: T) -> int:
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
+
         if key not in self._map:
             self._map[key] = len(self._items)
             self._items.append(key)
         return self._map[key]
 
     def update(self, sequence: SetLike[T]) -> int:
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
+
         item_index = 0
         for item in sequence:
             item_index = self.add(item)
@@ -954,6 +973,9 @@ class OrderedSet(StableSet[T]):
         return self._map[key]
 
     def pop(self, index: int = -1) -> T:
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
+
         if not self._items:
             raise KeyError("Set is empty")
         elem = self._items[index]
@@ -962,6 +984,9 @@ class OrderedSet(StableSet[T]):
         return elem
 
     def popitem(self, last: bool = True):
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
+
         if not self._items:
             raise KeyError("Set is empty")
         index = -1 if last else 0
@@ -971,6 +996,9 @@ class OrderedSet(StableSet[T]):
         return elem
 
     def move_to_end(self, key):
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
+
         if key in self:
             self.discard(key)
             self.add(key)
@@ -978,6 +1006,9 @@ class OrderedSet(StableSet[T]):
             raise KeyError(key)
 
     def discard(self, key: T) -> None:
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
+
         if key in self:
             i = self._map[key]
             del self._items[i]
@@ -995,6 +1026,9 @@ class OrderedSet(StableSet[T]):
         self._map = {item: idx for (idx, item) in enumerate(items)}
 
     def difference_update(self, *sets: SetLike[T]) -> None:
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
+
         items_to_remove = set()  # type: Set[T]
         for other in sets:
             items_as_set = set(other)  # type: Set[T]
@@ -1003,11 +1037,20 @@ class OrderedSet(StableSet[T]):
             [item for item in self._items if item not in items_to_remove]
         )
 
-    def intersection_update(self, other: SetLike[T]) -> None:
+    def intersection_update(self, other: SetLike[T]) -> T:
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
+
         other = set(other)
         self._update_items([item for item in self._items if item in other])
+        return self
+
+    __iand__ = intersection_update
 
     def symmetric_difference_update(self, other: SetLike[T]) -> None:
+        if self._is_mutable is False:
+            raise ValueError("This object is not mutable.")
+
         items_to_add = [item for item in other if item not in self]
         items_to_remove = set(other)
         self._update_items(
@@ -1018,6 +1061,8 @@ class OrderedSet(StableSet[T]):
 class SortedSet:
 
     def __init__(self, *args, set_=None, **kwargs):
+        self._is_mutable = True
+
         self._sorted = None
         if set_:
             self.set_ = set_
@@ -1060,6 +1105,9 @@ class SortedSet:
 
     def intersection_update(self, other):
         self.set_.intersection_update(other)
+        return self
+
+    __iand__ = intersection_update
 
     def __or__(self, other):
         # Union
@@ -1210,7 +1258,7 @@ class SortedSet:
         self.set_.remove(result)
         return result
 
-    def isorderedsubset(self: SetLike, other: SetLike, non_consecutive: bool = False):
+    def isorderedsubset(self: SetLike, other: SetLike, non_consecutive: bool = False) -> bool:
         if len(self) > len(other):
             return False
         if non_consecutive:
@@ -1228,5 +1276,5 @@ class SortedSet:
                     return False
             return True
 
-    def isorderedsuperset(self, other: SetLike, non_consecutive: bool = False):
+    def isorderedsuperset(self, other: SetLike, non_consecutive: bool = False) -> bool:
         return StableSet.isorderedsubset(other, self, non_consecutive)
